#include <HCSR04.h>
#include <SoftwareSerial.h>
#include "buzzer.h"

//#include "Arduino.h"

#define UP 1
#define DOWN 2
#define RIGHT 3
#define LEFT 4
#define NEUTRAL 0

/* PIN defines */
#define HCSR04_PIN_TRIG	7
#define HCSR04_1_PIN_ECHO	3
#define HCSR04_2_PIN_ECHO	4
#define HCSR04_3_PIN_ECHO	5
#define HCSR04_4_PIN_ECHO	6

#define LED_PIN 13
#define BUZZER_PIN 11

#define JOYSTICK_PIN_SW	2
#define JOYSTICK_PIN_VRX	A0
#define JOYSTICK_PIN_VRY	A1

#define DIGITAL_RX 8
#define DIGITAL_TX 9

const int HCSR04_CNT = 4;
const int echo_delay = 20;
const int no_reads = 2;

const int joystickDuration = 60000;
const int passwdDuration = 1000;
const int notificationDuration = 10000;

float avg1 = 0, avg2 = 0, avg3 = 0, avg4 = 0;
bool password_mode = false;
bool alert_mode = false;
int counter = 0;

unsigned long currentMillis = 0;
unsigned long previous_JOY_milis = 0;
unsigned long previousPasswdMillis = 0;
unsigned long notificationMillis = 0;
unsigned long previousBuzzerMillis = 0;
unsigned long previousBuzzerDelay = 0;
unsigned long previousLEDMillis = 0;

const int password_length = 4;
int i = 0;
int previous_move = 0;
const int password[] = {UP, UP, RIGHT, DOWN};
int ledState = 0; 
float notif_cnt = 0;

const int notes[] = {a, a, a, f, cH, a, f, cH, a, 0, eH, eH, eH, fH, cH, gS, f, cH, a, 0};
const int delays[] = {500, 500, 500, 350, 150, 500, 350, 150, 650, 500, 500, 500, 500, 350, 150, 500, 350, 150, 650, 500};

int index = 0;

SoftwareSerial espSerial(DIGITAL_RX, DIGITAL_TX);

HCSR04 hc(HCSR04_PIN_TRIG, 
          new int[HCSR04_CNT]{HCSR04_1_PIN_ECHO, 
                              HCSR04_2_PIN_ECHO, 
                              HCSR04_3_PIN_ECHO, 
                              HCSR04_4_PIN_ECHO},
          HCSR04_CNT); 

void send_notifs() {
  if (millis() - notificationMillis >= notificationDuration) {
    Serial.println("Notif sent!");
    espSerial.write(notif_cnt);
    notif_cnt++;    
    notificationMillis = millis();
  }
} 

void read_HCSR04() {
  float sum1 = 0, sum2 = 0, sum3 = 0, sum4 = 0;

  for (int i = 0; i < no_reads; i++) {
    sum1 += hc.dist(0);
    delay(echo_delay);
    sum2 += hc.dist(1);
    delay(echo_delay);
    sum3 += hc.dist(2);
    delay(echo_delay);
    sum4 += hc.dist(3);
    delay(echo_delay);
  }
  avg1 = sum1 / no_reads;
  avg2 = sum2 / no_reads;
  avg3 = sum3 / no_reads;
  avg4 = sum4 / no_reads;
}

void passwordMode() {
  digitalWrite(LED_PIN, 0);
  if (alert_mode)
    password_mode = true;
}

void check_password() {
  int input = input_password();
  if (input == NEUTRAL)
    return;
  if (input == previous_move) {
    previous_move = input;
    return; 
  }

  if (password[i] != -1 && password[i] != input) {
    i = 0;
  }
  if (password[i] == input && input != -1) {
    i++;
    if (i == password_length) {
      Serial.println("PASSWORD OK!");
      noTone(BUZZER_PIN);
      password_mode = false;
      alert_mode = false;
      i = 0;
    }
  }
}

int input_password() {
  int val_x = analogRead(JOYSTICK_PIN_VRX);
  int val_y = analogRead(JOYSTICK_PIN_VRY);
  
  Serial.println(val_x);
  Serial.println(val_y);

  if (val_x > -10 && val_x < 200)
    return UP;
  else if (val_x > 800 && val_x < 1100)
    return DOWN;
  else if (val_y > -10 && val_y < 200)
    return RIGHT;
  else if (val_y > 800 && val_y < 1100)
    return LEFT;
  else 
    return NEUTRAL;
}

void setup() {
  /* LED setup */
  pinMode(LED_PIN, OUTPUT);
  digitalWrite(LED_PIN, 0);
  
  /* Buzzer setup */
  pinMode(BUZZER_PIN, OUTPUT); 

  /* Serial setup */
  Serial.begin(9600);
  while (!Serial) {;}

  /* Joystick setup */
  pinMode(JOYSTICK_PIN_SW, INPUT_PULLUP);
  attachInterrupt(0, passwordMode, FALLING);
  password_mode = false;

  /* Digital serial setup - communication with ESP01 */
  espSerial.begin(115200);
}

void buzz() {
  if (currentMillis - previousBuzzerMillis >= delays[counter] ) {
      previousBuzzerMillis += delays[counter];
      tone(BUZZER_PIN, notes[counter]);
      digitalWrite(LED_PIN, 1);

    if (currentMillis - previousBuzzerDelay >= delays[counter] + 50) {
        previousBuzzerDelay += delays[counter] + 50;
        previousBuzzerMillis += 50;
        noTone(BUZZER_PIN);
        digitalWrite(LED_PIN, 0);
        counter++;
        if (counter == 20)
          counter = 0;
      }
  } 
}

void updateLedState() {
  int ledBlinkDuration = 100;
  int ledInterval = 100;
  if (ledState == 0) {
    if (currentMillis - previousLEDMillis >= ledInterval) {
       ledState = HIGH;
       previousLEDMillis += ledInterval;
    }
  }
  else {
    if (currentMillis - previousLEDMillis >= ledBlinkDuration) {
       ledState = LOW;
       previousLEDMillis += ledBlinkDuration;
    } 
  }    
}


void loop() {
  currentMillis = millis();

  if (alert_mode && password_mode) {
      buzz();
      updateLedState();
      digitalWrite(LED_PIN, ledState);
      if (currentMillis - previous_JOY_milis >= 200) {
        check_password();
        previous_JOY_milis += 200;
      }
  }

  if (alert_mode && !password_mode) {
    send_notifs();
    buzz();
    updateLedState();
    digitalWrite(LED_PIN, ledState);
  }

  if (!alert_mode && !password_mode) {
    read_HCSR04();

    if (avg1 < 50.0 || avg2 < 50.0 || avg3 < 50.0 || avg4 < 50.0) {
      alert_mode = true;
    }    
  }
  
}
